#### Functions implementing the RulEx-J model

## Argument definitions for all model functions
# cues: matrix of cue values, columns representing cues, rows representing trials
# judgments: judgments of the participant
# exemplars: matrix of exemplars, columns representing cues & the criterion value in the last column, rows representing exemplars 
# w: vector of weight parameters, first weight representing the constant [rule module]
# s: attention parameter [exemplar module]
# alpha: alpha parameter [RulEx-J model]
# parameters: vector of RulEx-J parameters, first element corresponding to alpha, last element to s, the remaining elements to w


## Rule module

# function to derive predictions
rule_module_pred <- function(w,cues) {
  predictions <- w[1]+rowSums(t(w[-1]*t(cues)))
  return(predictions)
}

# function to calculate difference between predictions and participants' judgments
rule_module_diff <- function(w,cues,judgments){
  predictions <- rule_module_pred(w,cues)
  diff <- predictions-judgments
  return(sum(diff^2))
}


## Exemplar module

# function to derive predictions
exemplar_module_pred <- function(s,exemplars,cues) {
  
  predictions <- rep(NA, nrow(cues))
  sim <- matrix(NA,nrow=nrow(exemplars),ncol=nrow(cues))
  
  for (i in 1:nrow(cues)){
    for (j in 1:nrow(exemplars)){
      sim[j,i] <- s^(4-sum(cues[i,]==exemplars[j,-ncol(exemplars)]))
    }
    predictions[i] <- sum(sim[,i]*t(exemplars[,ncol(exemplars)]))/sum(sim[,i])
  }

  return(predictions)
}

# function to calculate difference between predictions and participants' judgments
exemplar_module_diff <- function(s,exemplars,cues,judgments){
  predictions <- exemplar_module_pred(s,exemplars,cues)
  diff <- predictions-judgments
  return(sum(diff^2))
}


## RulEx-J model

# function to derive predictions
rulexj_pred<- function(alpha,w,s,exemplars,cues){
  pred_rule <- rule_module_pred(w,cues)
  pred_exemplar <- exemplar_module_pred(s,exemplars,cues)
  pred_rulexj <- alpha*pred_rule+(1-alpha)*pred_exemplar
  return(pred_rulexj)
}

# function to calculate difference between predictions and participants' judgments
# providing alpha, w, and s in separate arguments
rulexj_diff_sep <- function(alpha,w,s,exemplars,cues,judgments){
  predictions <- rulexj_pred(alpha,w,s,exemplars,cues)
  diff <- predictions-judgments
  return(sum(diff^2))
}


# function to calculate difference between predictions and participants' judgments
# providing alpha, w, and s as one argument (parameters)
# assuming alpha is the first and s the last parameter
rulexj_diff_joint <- function(parameters,exemplars,cues,judgments){
  alpha <- parameters[1]
  s <- parameters[length(parameters)]
  w <- parameters[2:(length(parameters)-1)]
  predictions <- rulexj_pred(alpha,w,s,exemplars,cues)
  diff <- predictions-judgments
  return(sum(diff^2))
}



#### Function to generate RulEx-J judgments

simulate_rulexj <- function(
  n, # number of participants to simulate
  cues, # cues for judgments
  exemplars, # exemplars
  alpha_fix=NULL, # optional parameter - if specified, alpha is kept constant at this value
  alpha_bounds=c(.00001,.99999), # bounds if alpha should be generated randomly
  w_bounds=c(0,100), # bounds for weight parameter generation
  s_bounds=c(.00,.50), # bounds for attention parameter generation
  w_sum_max=NULL # optional numeric value - if specified, the sum of all w cannot be greater than this value
){
  
  # Determine number of cues
  n_cues <- ncol(cues)
  
  # Prepare parameter matrix
  w_labels <- paste0('w',c(0:n_cues))
  parameters <- matrix(nrow=n,ncol=n_cues+4)
  colnames(parameters) <- c('ID','alpha',w_labels,'s')
  parameters[,'ID'] <- c(1:n)
  
  # Generate alpha parameter (or keep it fixed)
  if(is.null(alpha_fix)){
    parameters[,'alpha'] <- runif(n,min=alpha_bounds[1],max=alpha_bounds[2])
  } else {
    parameters[,'alpha'] <- alpha_fix
  }
  
  # Generate s parameter
  parameters[,'s'] <- runif(n,min=s_bounds[1],max=s_bounds[2])
  
  # Generate w parameters
  if (is.null(w_sum_max)){
    parameters[,w_labels] <- sapply(w_labels,function(x) runif(n,min=w_bounds[1],max=w_bounds[2]))
  } else {
    for (i in 1:nrow(parameters)){
      w_sum_above_max <- TRUE
      while(w_sum_above_max) {
        current_w <- runif(length(w_labels),min=w_bounds[1],max=w_bounds[2])
        w_sum_above_max <- sum(current_w)>w_sum_max
      }
      parameters[i,w_labels] <- current_w
    }
  }
  
  # Calculate judgments based on generated parameters
  judgments <- sapply(1:n, function(i){
    pars <- parameters[i,-1]
    judg <- rulexj_pred(alpha=pars['alpha'],w=pars[w_labels],s=pars['s'],exemplars=exemplars,cues=cues)
    return(cbind(i,judg,cues))
  },simplify = 'array')
  
  # Transform judgments array into matrix
  judgments <- apply(judgments,2,rbind)
  colnames(judgments) <- c('ID','judgment',paste0('c',c(1:n_cues)))
  
  
  return(list(parameters=parameters,judgments=judgments))
}


#### Function to fit RulEx-J

fit_rulexj <- function(
  dataset, # dataset containing the to be fitted judgments and cue patterns
  exemplars, # matrix of the exemplars stored in memory
  id_label='ID', # name of the column in the dataset containing the participant identifier
  cue_labels=c('c1','c2','c3','c4'), # vector of the column names in the dataset containing the cue patterns
  judgments_label='judgment', # name of the column in the dataset containing the judgments of the participant
  repetitions=10, # number of times the fitting procedure should be repeated with new randomly generated start parameters
  report=FALSE, # boolean indicating whether the function should report on its progress
  alpha_bounds=c(.00001,.99999), # lower and upper limit for the alpha parameter
  w_bounds=c(0,100), # lower and upper limit for the weight parameters
  s_bounds=c(.00,.50), # lower and upper limit for the attention parameter
  fit_rulexj_joint=TRUE # boolean indicating whether the alpha parameter should be fit jointly with the other parameters or afterwards
){
  
  # Extract number of cues and participant IDs
  n_cues <- length(cue_labels)
  ids <- unique(dataset[,id_label])
  
  
  # Prepare parameter labels
  
  parameters <- c(id_label,paste(c(paste0('w',0:n_cues),'SS_Rule','s','SS_Exemplar'),'sep',sep='_'))
  
  if (fit_rulexj_joint){
    parameters <- c(parameters,paste(c('alpha',paste0('w',0:n_cues),'s','SS_RulExJ'),'joint',sep='_'))
  } else {
    parameters <- c(parameters,'alpha_sep','SS_RulExJ_sep')
  }
  
  results <- matrix(NA,nrow=length(ids),ncol=length(parameters)) #generate empty results matrix
  colnames(results) <- parameters
  results <- data.frame(results)
  results[,id_label] <- ids
  
  # Fit models per participant
  for (ID in ids){
    
    # Filter data
    current_data <- dataset[dataset[,id_label]==ID,]
    cues <- as.matrix(current_data[,cue_labels])
    judgments <- as.vector(current_data[,judgments_label])
    
    # Set up results vector
    out <- c()
    
    
    # Fit rule module
    rule_start <- runif(n_cues+1,w_bounds[1],w_bounds[2])
    rule_fitted <- nlminb(start=rule_start, rule_module_diff,
                          cues=cues,judgments=judgments,
                          lower=rep(w_bounds[1],n_cues+1), upper=rep(w_bounds[2],n_cues+1))
    out <- c(out,rule_fitted$par,rule_fitted$objective)
    
    
    # Fit exemplar module
    
    # Repeat fitting process and collect results
    exemplar_fitted_SS <- rep(NA,repetitions)
    exemplar_fitted_s <- rep(NA,repetitions)
    
    for (rep in 1:repetitions){
      exemplar_start <- runif(1,s_bounds[1],s_bounds[2])
      exemplar_fitted <- nlminb(start=exemplar_start, exemplar_module_diff,
                                cues=cues,judgments=judgments,exemplars=exemplars,
                                lower=s_bounds[1], upper=s_bounds[2])
      exemplar_fitted_SS[rep] <- exemplar_fitted$objective
      exemplar_fitted_s[rep] <- exemplar_fitted$par
    }
    
    # Select best fitting parameters
    exemplar_fitted_SS[is.na(exemplar_fitted_SS)] <- Inf
    exemplar_fitted_best_SS <- min(exemplar_fitted_SS)
    exemplar_fitted_best_s <- exemplar_fitted_s[which.min(exemplar_fitted_SS)]
    out <- c(out,exemplar_fitted_best_s,exemplar_fitted_best_SS)
    
    
    #  Fit alpha parameter of RulEx-J model separately (if specified)
    
    if (fit_rulexj_joint==FALSE){
      # Repeat fitting process and collect results
      rulexj_fitted_SS <- rep(NA,repetitions)
      rulexj_fitted_alpha <- rep(NA,repetitions)
      
      for (rep in 1:repetitions){
        rulexj_start <- runif(1,alpha_bounds[1],alpha_bounds[2])
        rulexj_fitted <- nlminb(start=rulexj_start, rulexj_diff_sep,
                                  cues=cues,judgments=judgments,exemplars=exemplars,
                                  w=rule_fitted$par,s=exemplar_fitted_best_s,
                                  lower=alpha_bounds[1], upper=alpha_bounds[2])
        rulexj_fitted_SS[rep] <- rulexj_fitted$objective
        rulexj_fitted_alpha[rep] <- rulexj_fitted$par
      }
      
      # Select best fitting parameters
      rulexj_fitted_SS[is.na(rulexj_fitted_SS)] <- Inf
      rulexj_fitted_best_SS <- min(rulexj_fitted_SS)
      rulexj_fitted_best_alpha <- rulexj_fitted_alpha[which.min(rulexj_fitted_SS)]
      out <- c(out,rulexj_fitted_best_alpha,rulexj_fitted_best_SS)
    
    
    #  Fit parameters of RulEx-J model jointly (if specified)
    } else {
      
      # Set parameter bounds
      rulexj_lower <- c(alpha_bounds[1],rep(w_bounds[1],n_cues+1),s_bounds[1])
      rulexj_upper <- c(alpha_bounds[2],rep(w_bounds[2],n_cues+1),s_bounds[2])
                        
      # Repeat fitting process and collect results
      rulexj_fitted_SS <- rep(NA,repetitions)
      rulexj_fitted_parameters <- matrix(NA,nrow=repetitions,ncol=n_cues+3)
      
      for (rep in 1:repetitions){
        
        # optimization using nlminb works best
        rulexj_start <- runif(n_cues+3,rulexj_lower,rulexj_upper)
        
        rulexj_fitted <- nlminb(start=rulexj_start, rulexj_diff_joint,
                                cues=cues,judgments=judgments,exemplars=exemplars,
                                lower=rulexj_lower, upper=rulexj_upper,
                                control=list(eval.max=400,iter.max=300))
        
        rulexj_fitted_SS[rep] <- rulexj_fitted$objective
        rulexj_fitted_parameters[rep,] <- rulexj_fitted$par


      }
      
      # Select best fitting parameters
      rulexj_fitted_SS[is.na(rulexj_fitted_SS)] <- Inf
      rulexj_fitted_best_SS <- min(rulexj_fitted_SS)
      rulexj_fitted_best_parameters <- rulexj_fitted_parameters[which.min(rulexj_fitted_SS),]
      out <- c(out,rulexj_fitted_best_parameters,rulexj_fitted_best_SS)
      
    }
    
    # Store results
    results[results[,id_label]==ID,-1] <- out
    
    if(report){
      print(paste(id_label,ID,'finished'))  
    }
    
  }
  
  return(data.frame(results))
  
  
}

